
/* Portions Copyright (c) 1995,1996,1997 by Trusted Information Systems, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND TRUSTED INFORMATION SYSTEMS DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL TRUSTED INFORMATION 
 * SYSTEMS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 *
 * Trusted Information Systems, Inc. has received approval from the
 * United States Government for export and reexport of TIS/DNSSEC
 * software from the United States of America under the provisions of
 * the Export Administration Regulations (EAR) General Software Note
 * (GSN) license exception for mass market software.  Under the
 * provisions of this license, this software may be exported or
 * reexported to all destinations except for the embargoed countries of
 * Cuba, Iran, Iraq, Libya, North Korea, Sudan and Syria.  Any export
 * or reexport of TIS/DNSSEC software to the embargoed countries
 * requires additional, specific licensing approval from the United
 * States Government. 
 */




static char *rcsid = "$Header: /u/h/ogud/sec_bind/dnssec_lib/RCS/prandom.c,v 1.9 1996/11/20 22:07:57 ogud Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#ifdef SOLARIS
#include <sys/systeminfo.h>
#endif /* SOLARIS */
/*  This does not compile on the next machines I have access to */
  
#ifndef __P                /* Figure out what to do about prototypes */
#ifdef  __STDC__
#define __P(a) a
#else
#define __P(a) ()
#endif
#endif

static int  run_cmd       __P(( char *, char *, char *));
static void do_ls         __P(( char **));
static void get_random    __P(());
static void do_time       __P(());

/* local pointers to the variables passed into prandom  */
static void *prandom_ctx;
static int (*prandom_func) __P((void *, u_char *, int));

/* 
 * run_cmd runs the program with the arguments, if it finds it in the path,
 *  it is executed and the output is read from
 *  a pipe and fed into the digest function passed into prandom 
 * It returns the number of characters read from the pipe 
 */
static int 
run_cmd( name, args, path) 
char *name,  *args,  *path;
{
  /* allocate some memory for arument processing */ 
  char buffer[8192],      /* big enough for most commands output*/
       *s, command[256];  /* big enough for all system commands */
  int n, cnt=0;
  FILE *pipe;
  while( strlen(path) > 0) {
    s = strchr(path, ' ');                /* look for end of directory name */
    n = s ? (int) (s-path) : strlen(path);
    strncpy(command, path, n);            /* put directory name in command */
    command[n]='\0';
    sprintf (command, "%s/%s", command, name);  /* full name of commmand */
    if (!access(command, X_OK)) 
      break;
    while( s && *s == ' ') s++;                /* eat multiple spaces */
    path = s ? s : "";                  /* reseting the path */
  }
  if( *path == '\0')                      /* Error: command not found */
    return 0; 
                                          /* if it exists and is executable */
  sprintf( command, "%s %s 2>&1", command, args);
  pipe = popen( command, "r");            /* execute the command */
  while ((n=fread (buffer, sizeof(char), sizeof(buffer), pipe))>0){
    prandom_func(prandom_ctx, buffer, n); /* process the output */
    cnt+=n;
  }
  n = pclose(pipe);    /*  pclose() returns the exit code of program */
  if( n ) {
    printf("pclose() Returned %d %s\n",n, command);
    return -n;         /* if error do not return count indicate error */
  }
  return cnt;  
}

/* this function simulates the ls command, but it uses stat which gives more
 * information and is harder to guess 
 * do_ls() does not visit subdirectories
 */
static void 
do_ls( dirs)
char **dirs;
{
  int i,n=0;
  struct stat buf;
  DIR *dir;
  struct dirent *entry;
  for (i=0; dirs[i] && (strlen(dirs[i])>0);i++){/*for all directores on path*/
    if ((dir = opendir(dirs[i]))){              /* open it for read */
      while ((entry = readdir(dir))) {          /* for all entries in dir */
	stat(entry->d_name, &buf);              /* get the stats */
	prandom_func(prandom_ctx, (u_char *) &buf, sizeof(struct stat));  /* update */
	n++;
      }
      closedir(dir);                            /* done */
    }
  }
}

/* 
 * If /dev/random is available read some data from it. 
 */
static void 
get_random() {
  int n, fd;
  u_char buf[16];
  if (!access( "/dev/random", R_OK)) {
    fd = open( "/dev/random", O_RDONLY);
    n = read(fd, buf, sizeof(buf));
    prandom_func( prandom_ctx, buf, n);
    close(fd);
  }
}

/* Portable way of getting the time values if gettimeofday is missing 
 * then compile with -DMISSING_GETTIMEOFDAY  time() is POSIX compliant but
 * gettimeofday() is not.
*/ 
static void 
do_time()  
{
#ifdef MISSING_GETTIMEOFDAY
    time_t mtime; 
    mtime= time();
#else
    struct timeval mtime;
    struct timezone  zone;
    gettimeofday( &mtime, &zone);
    prandom_func( prandom_ctx, (u_char *) &zone, sizeof(zone));
#endif
    prandom_func( prandom_ctx, (u_char *) &mtime, sizeof(mtime));
  }

static void
get_ids()
{
  int temp;
  temp = (int) getpid(); /* maybe (?) hard to predict,
                               * limited range 0..32K*/ 
  prandom_func( prandom_ctx, (u_char *) &temp, sizeof(temp));
  
  temp = (int) getuid(); /* predicable limited range */ 
  prandom_func( prandom_ctx, (u_char *) &temp, sizeof(temp));
  
  temp = (int) getppid();/* easy to predict range 0..32K */ 
  prandom_func( prandom_ctx, (u_char *) &temp, sizeof(temp));
#ifndef SOLARIS  
  temp = (int) gethostid(); /* not unpredictable at all, but oh well */
  prandom_func( prandom_ctx, (u_char *) &temp, sizeof(temp));
#else 
  sysinfo(SI_HW_SERIAL, str, (SEED_LEN*4)-1);
  prandom_funct (prandom_ctx, (unsigned char *) str, strlen(str));
#endif
}
/*
 * common Directories that change frequently, can add to this list 
 */
static char *dirs[] = {"/tmp", "/usr/tmp", "/var/tmp", "/usr/adm",
		"/usr/spool", "/usr/mail", "/dev", "/var/spool",
		"/var/spool/mail", "/var/mail", NULL};
void 
prandom( A, func)
void *A;
int (*func) __P((void *, u_char *, int));
{
  prandom_ctx = A;
  prandom_func = func;
  do_time();

/* ps is a hard command to make protable  error messages are sometimes longer
 * than valid output in that case just run them all 
 */ 
  (run_cmd(  "ps", "-alxw", "/bin") > 300) || /* bsd and gnu ps all proc */
   (run_cmd( "ps", "-ef",   "/bin") > 300) || /* sysv ps most proc */
    (run_cmd("ps", "-ale",  "/bin") > 300) || /* portable, some no root proc */
      printf("prandom():PS not found\n");
	
  run_cmd( "df", "", "/bin /usr/sbin") ||   /* df either exists or not */
      printf("prandom():DF not found\n");
#ifndef __linux__
  run_cmd( "iostat", "", "/bin /usr/bin /usr/sbin") || 
      printf("prandom():iostat not found\n");
#endif 
  run_cmd(  "vmstat", "", "/usr/ucb /usr/bin /usr/sbin") ||
    run_cmd("vm_stat", "", "/usr/ucb /usr/bin") ||
      printf("prandom():vmstat and vm_stat not found\n");

  get_ids();

  run_cmd( "netstat", "-an", "/usr/ucb /usr/bin /usr/etc /usr/sbin /bin") ||
      printf("prandom():netstat not found\n");
#ifndef __linux__
  run_cmd( "netstat", "-s",  "/usr/ucb /usr/bin /usr/etc /usr/sbin /bin") ||
      printf("prandom():netstat not found\n");
#endif

  run_cmd( "arp", "-a", "/usr/etc /etc /usr/sbin /usr/bin /sbin") ||
      printf("prandom():arp not found\n");

  do_ls(dirs);

  /* W gives more output than uptime, and it is harder to predict but
     it is not allways available */
  run_cmd(  "w", "", "/usr/ucb /usr/bin /usr/bsd") ||
    run_cmd("uptime", "", "/usr/ucb /usr/bin /usr/bsd") ||
      printf("prandom():UPTIME and W not found\n");

  run_cmd( "printenv", "", "/usr/ucb /usr/bin /bin")  ||
      printf("prandom():PRINTENV not found\n"); 

  get_random();

  run_cmd( "hostname", "", "/bin /usr/bin /usr/ucb")  ||
      printf("prandom():HOSTNAME not found\n");

  do_time(); /* do time again this is harder to guess than the first one */

  return;
}
